#* This optimizing procedure is from the book numerical recipes in C Oxford Press
# Constrained flexible polyhedron algorithm 
# 19.02.2005 changed function value parser to parse also cost = (val_real_db,val_ph) values for .ac simulation
# with ltspice .meas ( see .. sub getfuncval ~ line 375 )
# 21.02.2005 changed line 200 $mpts=$ndim+1 ( thanks to gerd for error reporting )
# 23.02.2005 added debug option ( debug=0,1,2 in the init-file 0=nodebug 1=parametervalues 2=full debug ) to display additional info ( parametervalues , geometric operation ) at each iteration
#            automatically convert varnames to lowercase before testing of equality
#
# 24.02.2005 added beep at end optimization
#            added absftol=value option intifile to stop optimization if funcval<=absftol
#            added bestfuncval output at each iteration (bestvalue so far)
# 25.02.2005 corrected output error in line 301 "Can't open source file : $initfile->$file\n";
#            added to read_deck()   s/\s*$//;  eats blanks at end of line (sometimes caused troubles)
#            change line 547 to  if $debug>=1
# 06.03.2005 add yb=1e100, add some checking of consistence of initfile see about line 250
#
# 26.05.2005 some code modifications to handle log(0) error messages (sort,fluct,amotry)
#            some code modifications to handle -1#INF pattern from ltspice measure command (funcval=0 in ac analysis) 
#            (get_funcval)
$file="";
$simendfile="tinyperl simend.pl";
$ascfile="myfile.asc";
$rtol=0;
$ftol=0.2;
$tiny=1e-20;
$delt=5;
$itermax=100;
$ndim=2;
$mpts=$ndim+1;
$lolimits='0 0';
$uplimits='1 1';
$initvals='0 0';
@lolim=();
@uplim=();
@initval=();
@ptry = (); # dim: ndim
@y = ();    # dim: mpts
@psum = (); # dim: ndim
@p = (); # dim: mpts,ndim
$ihi=2;
$inhi=1;
$ilo=0;
$debug=0;
$absftol=0;
$bestfuncval=1e100;
$eps=1e-7;
# new for metropolis algorithm ( simulating annealing ) -> experimental
$idum=1;
$ylo=0;
$yhi=0;
$ynhi=0;
$yt=0;
$yb=1e100;
$yflu=0;
@pb=(); # dim: ndim
$temptr=1;
$tt=-$temptr;
$epsilon=0.9;
$nr_temp_iter=4;
# new end

$funcval=0;
@funcvals = (); # dim: maxiter+1
@varval=();
@varname=();
$all_pm;
@deck=();
$funcfile="myfile.log";
$initfile="myfile_init.txt";
$cirfile="myfile.cir";
$varnames="{var1}_{var2}";
if(!($ARGV[0] eq ''))  { $initfile=$ARGV[0]; }
$spicecmdline="ltspice -b ";
print "\noptimizer with flexible polyhedron alg. combined with thermal fluctuations\n";
print "from the book : numerical recipes in C - second edition - Cambridge University Press\n\n";
print "starting optimization .....";
$cnt=0;
&init;
&main;
sub main
{

my $i,$j,$temp,@argu,$ysave;

$mpts=$ndim+1;
while ($cnt<$itermax)
{
	&sort;
	&fluct;
	$rtol=2*(abs($yhi-$ylo))/(abs($yhi)+abs($ylo)+ $tiny);
	if($debug==2)
	{
	    print "\n    rtol=$rtol  ftol=$ftol  best_funcval_so_far=$bestfuncval\n";
	}
	if ($rtol<$ftol)
	{
		$swap=$y[0];
		$y[0]=$y[$ilo];
		$y[$ilo]=$swap;
		for($n=0;$n<$ndim;$n++)
		{
		   $swap=$p[0][$n];
		   $p[0][$n]=$p[$ilo][$n];
		   $p[$ilo][$n]=$swap;
		}
		goto ende; # we are done now
	}
	if ($bestfuncval<$absftol)
	{
		$swap=$y[0];
		$y[0]=$y[$ilo];
		$y[$ilo]=$swap;
		for($n=0;$n<$ndim;$n++)
		{
		   $swap=$p[0][$n];
		   $p[0][$n]=$p[$ilo][$n];
		   $p[$ilo][$n]=$swap;
		}
		goto ende; # we are done now
	}
	&amotry(-1); # do reflection of simplex
	if ($ytry <= $ylo)
	{
	    &amotry(2); # if now the best point try additional expansion by factor 2
	}
	else
	{
		if ($ytry >= $ynhi)
		{
			$ysave=$yhi;
			&amotry(0.5); # if not better than the 2nd worst do 1dim contraction
			if ($ytry>=$ysave) # if now even the worst , you can only do multiple contraction of all points -> directed to the best point
			{
				$i=0;
				print "\ndo multiple contraction to best point ..............................\n";
				if($debug==2)
                {
                	print "\nbest point:\n\n";
               		for($j=0;$j<$ndim;$j++)
               		{
                  		print "    p[$ilo][$j]=$p[$ilo][$j]  \n";
               		}
               	}
                print "\n";
				while ($i < $mpts)
				{ 	if ($i != $ilo)
				    {
				    	if($debug==2)
                        {  # print worst point
                           print "\n    point before:\n\n";
                           for($j=0;$j<$ndim;$j++)
                           {
                              print "    p[$i][$j]=$p[$i][$j]  \n";
                           }
                           print "\n";
                        }
				    	for($j=0;$j<$ndim;$j++)
   				        {
   				            $p[$i][$j]=0.5*($p[$i][$j]+$p[$ilo][$j]);
                            $psum[$j]=$p[$i][$j];
                            $ptry[$j]=$psum[$j];
                        }
                        if($debug>=1)
                        {  # print actual ptry values
                           print "\n    new try:\n\n";
                           for($j=0;$j<$ndim;$j++)
                           {
                              print "    ptry[$j]=$ptry[$j]  \n";
                           }
                           print "\n";
                        }
                        &func;
                        $y[$i]=$funcval;
					}
					$i=$i+1;
				}
				print "\nmultiple contraction to best point ended ..............................\n";
				&cpt_psum;
			 }
		}
	}
}
ende:
print "\noptimization ended !!!! \n";
print "\a";

print "\nbest point so far:\n\n";

for($j=0;$j<$ndim;$j++)
{
	  print "pbest[$j]=$pb[$j]";
	  print "\n";
}

print "\nbest funcval = $bestfuncval\n";

$varvals="";
for($j=0;$j<$ndim;$j++) # concatenate varvals to string
{
    $varvals=$varvals.$pb[$j]."_";
}
chop($varvals);
print "\nloading ltspice in graphics mode .... ";
print "\n\n( to exit the optimizer please close the current ltspice session )\n\n";
@argu = ("$simendfile $ascfile $varvals $varnames");
system(@argu) == 0
     or die "system @args failed: $?";
print "bye :-) \n\n";
exit(0);
}



sub init
{
# init initial simplex points by user
# p(anzpt=anzpm+1,anzpm) , psum(anzpm), y(anzpt) , ptry(ndim, funcvals(maxiter) = vectors
# funcval = skalar
	my ($i,$k,$j,$m,$idx,$rev,$delt,$linenum,@delta);

	$file=$initfile;
	print "\nreading initfile ..... ";
	&read_deck;

	for ($i=0;$i<@deck;$i++)    # for the whole deck
    {
       $_=$deck[$i];
       if(m/\bftol\b\s*\=(.*)/i) {$ftol=$1;}
       if(m/\bdelt\b\s*\=(.*)/i) {$delt=$1;}
       # no more evaluated: if(m/\btiny\b\s*\=(.*)/i) {$tiny=$1;}
       if(m/\bndim\b\s*\=(.*)/i) {$ndim=$1;}
       if(m/\bitermax\b\s*\=(.*)/i) {$itermax=$1;}
       if(m/\bfuncfile\b\s*\=(.*)/i) {$funcfile=$1;}
       if(m/\bspicecmdline\b\s*\=(.*)/i) {$spicecmdline=$1;}
       if(m/\bcirfile\b\s*\=(.*)/i) {$cirfile=$1;}
       if(m/\bsimendfile\b\s*\=(.*)/i) {$simendfile=$1;}
       if(m/\bascfile\b\s*\=(.*)/i) {$ascfile=$1;}
       if(m/\blolimits\b\s*\=(.*)/i) {$lolimits=$1;}
       if(m/\buplimits\b\s*\=(.*)/i) {$uplimits=$1;}
       if(m/\binitvals\b\s*\=(.*)/i) {$initvals=$1;}
       if(m/\bvarnames\b\s*\=(.*)/i) {$varnames=$1;}
       if(m/\bidum\b\s*\=(.*)/i) {$idum=$1;}
       if(m/\btemptr\b\s*\=(.*)/i) {$temptr=$1;}
       if(m/\bepsilon\b\s*\=(.*)/i) {$epsilon=$1;}
       if(m/\bnr_temp_iter\b\s*\=(.*)/i) {$nr_temp_iter=$1;}
       # no more evaluated: if(m/\byb\b\s*\=(.*)/i) {$yb=$1;}
       if(m/\bdebug\b\s*\=(.*)/i) {$debug=$1;}
       if(m/\babsftol\b\s*\=(.*)/i) {$absftol=$1;}
    }

	$tt=-$temptr;
    $varnames=lc($varnames);
	$all_pm=@lolim=split(/\s+/,$lolimits);
	if($all_pm!=$ndim) {print "error: inconstistent count of parameternames versus lolim values detected -> check lolim !";exit(0)}
	$all_pm=@uplim=split(/\s+/,$uplimits);
	if($all_pm!=$ndim) {print "error: inconstistent count of parameternames versus uplim values detected -> check uplim !";exit(0)}
	$all_pm=@initval=split(/\s+/,$initvals);
	if($all_pm!=$ndim) {print "error: inconstistent count of parameternames versus initval values detected -> check initval !";exit(0)}
	@varname=split(/\_/,$varnames);
    $all_pm=@varname;
    if($all_pm!=$ndim) {print "error: inconstistent count of parameternames versus ndim value detected -> check ndim !";exit(0)}
    $mpts=$ndim+1; # dont forget to reinitialize it !!!!!!
	@deck=(); # now free
	$file=$cirfile;
    # read in circuitdeck !!!! ( needed by sub func )
    print "\nreading circuitfile ..... \n\n";
    &read_deck;

	# generate initsimplex from initvals and limits

	for($i=0;$i<$ndim;$i++)
	{
	   $delta[$i]=($uplim[$i]-$lolim[$i])/$delt;
	}
	for($i=0;$i<$mpts;$i++) # one point of initial simplex can be userdefined , copy it to the others = default
	{
	   for($j=0;$j<$ndim;$j++)
	   {
	        $p[$i][$j]=$initval[$j];
	   }
	}
	# now change the other points ( not the user defined point )
	$k=0;$m=0;$idx=1+$mpts*$k;
	$rev=0;
	for($i=1;$i<$mpts;$i++) # the other points of the initial simplex
	{
	    for($j=0;$j<$ndim;$j++)
	    {
	        $m++;
	        if($m!=$idx)
	        {
	           $p[$i][$j]=$p[$i][$j]+$delta[$j];
	           if($p[$i][$j]>$uplim[$j]) {$rev=1;}
	        }
	        else
	        {
	           $k++;$idx=1+$mpts*$k;
	        }
	    }
	}
	if($rev==1) # seams better with subtraction of delta -> instead of addition
	{
	   $k=0;$m=0;$idx=1+$mpts*$k;
	   for($i=1;$i<$mpts;$i++)
	   {
	      for($j=0;$j<$ndim;$j++)
	      {
	        $m++;
	        if($m!=$idx)
	        {
	           $p[$i][$j]=$p[$i][$j]-2*$delta[$j];
	        }
	        else
	        {
	           $k++;$idx=1+$mpts*$k;
	        }
	      }
	   }
	}
    # get the function values for the initial simplex
    &cpt_psum;
    print "\ninital simplex points:\n\n";
	for($i=0;$i<$mpts;$i++)
	{
	   for($j=0;$j<$ndim;$j++)
       {
               $ptry[$j]=$p[$i][$j];
       }
       if($debug>=1)
       {  # print actual ptry values
           print "\n";
           for($j=0;$j<$ndim;$j++)
           {
              print "     ptry[$j]=$ptry[$j]  ";
           }
           print "\n\n";
       }
	   &func;
	   $y[$i]=$funcval;
	}
	print "\n\nsimplex initialization ended. Now starting geometric operations on simplex:\n\n";
}



sub read_deck
{
    open (FILE,$file) || die "Can't open source file : $file\n";
    $_=<FILE>;chop;@deck="" . $_;   # heading/comment line passthrough

    while (<FILE>)
    {
      chomp; # chomp is better than chop -> last line .end -> .en error
      # dont lowercase stuff in quotes
      $linenum = $.; # act. linenumber
      if( not $linenum eq 1 ) # let the first line unchanged
      {
         #if (/(.*)([\'\"])(.*)([\"\'])(.*)/)
         #{
         #     $_=lc($1) . $2 . $3 . $4 . lc($5);
         #}
         $_ = lc($_);
         s/\$.*//; # no trailing comments with $
         s/\;.*//; # no trailing comments with ;
         s/^\*.*//;  # no comments starting with *
         s/^\s*$//;  # no bl lines
         s/\s*$//;  # eats blanks at end of line (sometimes caused troubles)
         if (/^\s*\+(.+)/)
         { # continuation
              $_ = pop (@deck) . " " . $1;
         }
      }
      push @deck,$_ if (length($_) >0);
    }
    close(FILE);
}


sub func
{
    my ($test,$j,@argu,$go,$anz_pm,$n,$varvals);

    for($j=0;$j<$ndim;$j++)
    {
    	if ($ptry[$j]<$lolim[$j])
    	{
    	    if($debug==2) {print "\n    ptry[$j] ($ptry[$j]) lower than lolim -> limit to lolim ($lolim[$j])\n\n\n";}
    	    $ptry[$j]=$lolim[$j];
    	}
    	if ($ptry[$j]>$uplim[$j])
    	{
    	    if($debug==2) {print "\n    ptry[$j] ($ptry[$j]) higher than uplim -> limit to uplim ($uplim[$j])\n\n\n";}
    	    $ptry[$j]=$uplim[$j];
    	}
    }
    $varvals="";
    for($j=0;$j<$ndim;$j++) # concatenate varvals to string
    {
    	$varvals=$varvals.$ptry[$j]." ";
    }
    @varval=split(/\s+/,$varvals);
    # now update parameter of loaded cirfile ( copydeck )
    @copydeck=@deck;
    $go=1;
    $j=0;
    $anz_pm=0;
    while ($go)
    {
      $j++;
      $_=$copydeck[$j];
      for($n=0;$n<$all_pm;$n++)
      {
         # added ignore case switch /i
         if(s/$varname[$n]/$varval[$n]/i) # {x1} -> val1    {x2} -> val2
         {
            $anz_pm++;
            $copydeck[$j]=$_; # update
            if($anz_pm==$all_pm)
            {$go=0;} # we are done now
         }
      }
    }
    # write deck to cirfile.tmp and simulate it
    $_=$cirfile;
    s/\.(.*)/.tmp/; # cirfile.cir or cirfile.net or cirfile.sp to file.tmp
    $outfile=$_;
    # write the changed file to cirfile.tmp
    &fprintdeck;


    # run simulation with new parameters
    @argu = ("$spicecmdline $outfile");
        system(@argu) == 0
              or die "system @args failed: $?";
    #######################
    $funcval=&get_funcval;
    #######################
    if($funcval<$bestfuncval) {$bestfuncval=$funcval;}
    $funcvals[$cnt]=$funcval;
    if ($funcval<$yb) # init yb to a high value to enable saving
    {
        for($j=0;$j<$ndim;$j++) # save best vector
        {
            $pb[$j]=$ptry[$j];
        }
        $yb=$funcval;
    }
    print "iteration $cnt: funcvalue=$funcval ($bestfuncval)\n";
    $test=$cnt%$nr_temp_iter; # change templevel every nr_temp_iter iteration
	if(!($test)) # alter temp level for simulation annealing addition
    {
       $tt=$tt*(1-$epsilon);
       if($tt)
       {
          print "\n =========================================";
          print "\n     actual temperature=$tt";
          print "\n =========================================\n\n";
       }
    }
    $cnt++;
}

sub get_funcval
{
    # gets funcval from file
    # open()
    open (LOGFILE,$funcfile) || die "Can't open source file : $funcfile\n";
    $_=<LOGFILE>;chomp;
    $go=1;
    while($go)
    {
      $_=<LOGFILE>;
      if(m/\bcost\b(.*?)=\s*(\S+)/) # search for cost ... =  val
      {
          $_=$fval=$2;

	  if(m/\(([^\,]+)/)   # \,([^\)])+/)  search for ( .... , to read  ($fval_mag_db,  $fval_phase_deg) expressions
	  {
		$fval_mag=$1; # get ....
		$fval=substr($fval_mag,0,-2); # delete db suffix
		if($fval=='-1.#INF') 
		{$fval=0;}
            else
            {$fval=10**($fval/20);}  # compute real-value from db
        }
	  $funcval=$fval;
	  #$_=<LOGFILE>;$_=<LOGFILE>;
	  #print "\n$_\n";
	  #print "\n\nfuncval from deck: $funcval\n\n";
	  $go=0;
      }
    }
    close(LOGFILE);
    return $funcval;
}

sub cpt_psum
{   # get the sum of all p[0][x] p[1][x] ... p[mpts-1][x]
    #                    p[0][y] p[1][y] ... p[mpts-1][y]
    local $j,$i,$sum;

    $j=0;
    while ($j<$ndim)
    {
        $sum=0;
        $i=0;
        while ($i<$mpts)
        {
            $sum=$sum+$p[$i][$j];
            $i=$i+1;
        }
        $psum[$j]=$sum;
        $j=$j+1;
    }
}

sub sort
{
    local $i;
    #computes ihi inhi ilo according to yhi ynhi ylo
    $ilo=0;
    $ihi=1;
    $ynhi=$ylo=$y[0]+$tt*log($eps+rand($idum));  # only for the simulating annealing part
    $i=0;
    while ($i<$mpts)
    {
        if ($y[$i]<=$y[$ilo])
        { $ilo=$i; }
        if ($y[$i]>$y[$ihi])
        {$inhi=$ihi;$ihi=$i; }
        else
        {
            if ($y[$i]>$y[$inhi])
            {
                if ($i!=$ihi)
                {
                    $inhi=$i;
                }
            }
        }
        $i=$i+1;
    }
}


sub fluct # only needed for simulting annealing part of optimization
{
    for($i=0;$i<$mpts;$i++)
    {
       $yt=$y[$i]+$tt*log($eps+rand($idum)); # add temperature
       if($yt<=$ylo)
       {
          $ilo=$i;
          $ylo=$yt;
       }
       if($yt>$yhi)
       {
          $ynhi=$yhi;
          $ihi=$i;
          $yhi=$yt;
       }
       else
       {
          if($yt>$ynhi)
          {
             $ynhi=$yt;
          }
       }
    }
}

sub amotry # do all the geometric operation on the n-dim-simplex e.g. reflection,contraction(1dim),contraction(multidim),expansion
{
    local $j,$fac1,$fac2,$fac;

    $fac = $_[0]; # the function parameter = 0.5 or 2 or -1
    if($debug==2)
    {
    	if($fac==0.5) {print "\n    contraction of worst point\n";}
    	if($fac==-1.0) {print "\n    reflection of worst point\n";}
    	if($fac==2.0) {print "\n    expansion of now best point\n";}
    }
    $fac1 = (1-$fac)/$ndim;
    $fac2 = $fac1-$fac;
    for($j=0;$j<$ndim;$j++)
    {
         $ptry[$j] = $psum[$j]*$fac1-$p[$ihi][$j]*$fac2;
    }
    if($debug==2)
    {
       print "\nbest point:\n\n";
       for($j=0;$j<$ndim;$j++)
       {
         	print "    p[$ilo][$j]=$p[$ilo][$j]  \n";
       }
       # print worst point
       print "\n    point before:\n\n";
       for($j=0;$j<$ndim;$j++)
       {
          print "    p[$ihi][$j]=$p[$ihi][$j]  \n";
       }
       print "\n";
    }
    if($debug>=1)
    {  # print actual ptry values
       print "\n    new try:\n\n";
       for($j=0;$j<$ndim;$j++)
       {
          print "    ptry[$j]=$ptry[$j]  \n";
       }
       print "\n";
    }
    &func; # calls the function evaluation routine
    $ytry = $funcval;
    $yflu=$ytry-$tt*log($eps+rand($idum)); # only for the simulating annealing part
    if ($yflu<$yhi)
    {
        $y[$ihi]=$ytry;
        $yhi=$yflu;
        for($j=0;$j<$ndim;$j++)
        {
            $psum[$j]=$psum[$j]+$ptry[$j]-$p[$ihi][$j];
            $p[$ihi][$j]=$ptry[$j];
        }
    }
    $ytry=$yflu;
}

sub fprintdeck {

  open (DATEI,"+>$outfile") || die "Can't open source file : $outfile\n";
  for (@copydeck) {
    print DATEI $_."\n";
  }
  close (DATEI);
}